
/*
 * New grid file format for VIS-5D:
 *
 * The header is a list of tagged items.  Each item has 3 parts:
 *    1. A tag which is a 4-byte integer identifying the type of item.
 *    2. A 4-byte integer indicating how many bytes of data follow.
 *    3. The binary data.
 *
 * If we need to add new information to a file header we just create a
 * new tag and add the code to read/write the information.
 *
 * If we're reading a header and find an unknown tag, we can use the
 * length field to skip ahead to the next tag.  Therefore, the file
 * format is forward (and backward) compatible.
 *
 * Grid data is stored as either:
 *     1-byte unsigned integers  (255=missing)
 *     2-byte unsigned integers  (65535=missing)
 *     4-byte IEEE floats        ( >1.0e30 = missing)
 *
 * All numeric values are stored in big endian order.  All floating point
 * values are in IEEE format.
 */



/*
 * Updates:
 *
 * April 13, 1995, brianp
 *   finished Cray support for 2-byte and 4-byte compress modes
 */




#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "binio.h"
#include "v5d.h"
#include "vis5d.h"
#ifndef SEEK_SET
#  define SEEK_SET 0
#endif
#ifndef SEEK_CUR
#  define SEEK_CUR 1
#endif
#ifndef SEEK_END
#  define SEEK_END 2
#endif
/*
 * Currently defined tags:
 * Note:  the notation a[i] doesn't mean a is an array of i elements,
 * rather it just refers to the ith element of a[].
 *
 * Tags marked as PHASED OUT should be readable but are no longer written.
 * Old tag numbers can't be reused!
 *
 */


/*      TAG NAME        VALUE       DATA (comments)                     */
/*----------------------------------------------------------------------*/
#define TAG_ID          0x5635440a  /* hex encoding of "V5D\n"          */

/* general stuff 1000+ */
#define TAG_VERSION     1000        /* char*10 FileVersion              */
#define TAG_NUMTIMES    1001        /* int*4 NumTimes                   */
#define TAG_NUMVARS     1002        /* int*4 NumVars                    */
#define TAG_VARNAME     1003        /* int*4 var; char*10 VarName[var]  */

#define TAG_NR          1004        /* int*4 Nr                         */
#define TAG_NC          1005        /* int*4 Nc                         */
#define TAG_NL          1006        /* int*4 Nl  (Nl for all vars)      */
#define TAG_NL_VAR      1007        /* int*4 var; int*4 Nl[var]         */
#define TAG_LOWLEV_VAR  1008        /* int*4 var; int*4 LowLev[var]     */

#define TAG_TIME        1010        /* int*4 t;  int*4 TimeStamp[t]     */
#define TAG_DATE        1011        /* int*4 t;  int*4 DateStamp[t]     */

#define TAG_MINVAL      1012        /* int*4 var;  real*4 MinVal[var]   */
#define TAG_MAXVAL      1013        /* int*4 var;  real*4 MaxVal[var]   */

#define TAG_COMPRESS    1014        /* int*4 CompressMode; (#bytes/grid)*/

#define TAG_UNITS       1015        /* int *4 var; char*20 Units[var]   */

/* vertical coordinate system 2000+ */
#define TAG_VERTICAL_SYSTEM 2000    /* int*4 VerticalSystem             */
#define TAG_VERT_ARGS    2100       /* int*4 n;  real*4 VertArgs[0..n-1]*/

#define TAG_BOTTOMBOUND  2001       /* real*4 BottomBound     (PHASED OUT)  */
#define TAG_LEVINC       2002       /* real*4 LevInc      (PHASED OUT)      */
#define TAG_HEIGHT       2003    /* int*4 l;  real*4 Height[l] (PHASED OUT) */


/* projection 3000+ */
#define TAG_PROJECTION   3000       /* int*4 projection:                    */
                                    /*   0 = generic linear                 */
                                    /*   1 = cylindrical equidistant        */
                                    /*   2 = Lambert conformal/Polar Stereo */
                                    /*   3 = rotated equidistant            */
#define TAG_PROJ_ARGS    3100       /* int *4 n;  real*4 ProjArgs[0..n-1]   */

#define TAG_NORTHBOUND   3001       /* real*4 NorthBound       (PHASED OUT) */
#define TAG_WESTBOUND    3002       /* real*4 WestBound        (PHASED OUT) */
#define TAG_ROWINC       3003       /* real*4 RowInc           (PHASED OUT) */
#define TAG_COLINC       3004       /* real*4 ColInc           (PHASED OUT) */
#define TAG_LAT1         3005       /* real*4 Lat1             (PHASED OUT) */
#define TAG_LAT2         3006       /* real*4 Lat2             (PHASED OUT) */
#define TAG_POLE_ROW     3007       /* real*4 PoleRow          (PHASED OUT) */
#define TAG_POLE_COL     3008       /* real*4 PoleCol          (PHASED OUT) */
#define TAG_CENTLON      3009       /* real*4 CentralLon       (PHASED OUT) */
#define TAG_CENTLAT      3010       /* real*4 CentralLat       (PHASED OUT) */
#define TAG_CENTROW      3011       /* real*4 CentralRow       (PHASED OUT) */
#define TAG_CENTCOL      3012       /* real*4 CentralCol       (PHASED OUT) */
#define TAG_ROTATION     3013       /* real*4 Rotation         (PHASED OUT) */


#define TAG_END                9999

#define FILE_VERSION "4.2a"




/*
 * Open a v5d file for reading.
 * Input:  filename - name of v5d file to open
 *         v - pointer to a v5dstruct in which to put header info or NULL
 *             if a struct should be dynamically allocated.
 * Return:  NULL if error, else v or a pointer to a new v5dstruct if v was NULL
 */
v5dstruct *v5dOpenFile( const char *filename, v5dstruct *v )
{
   int fd;

   fd = open( filename, O_RDONLY );
   if (fd==-1) {
      /* error */
      return 0;
   }

   if (v) {
      v5dInitStruct( v );
   }
   else {
      v = v5dNewStruct();
      if (!v) {
         return NULL;
      }
   }

   v->FileDesc = fd;
   v->Mode = 'r';
   if (read_v5d_header( v )) {
      return v;
   }
   else {
      return NULL;
   }
}

/*
 * Return current file position.
 * Input:  f - file descriptor
 */
static off_t ltell( int f )
{
   return lseek( f, 0, SEEK_CUR );
}



/*
 * Return a pointer to a new, initialized v5dstruct.
 */
v5dstruct *v5dNewStruct( void )
{
   v5dstruct *v;

   v = (v5dstruct *) malloc( sizeof(v5dstruct) );
   if (v) {
      v5dInitStruct(v);
   }
   return v;
}

/*
 * Initialize a v5dstructure to reasonable initial values.
 * Input:  v - pointer to v5dstruct.
 */
void v5dInitStruct( v5dstruct *v )
{
   int i;

   /* set everything to zero */
   memset( v, 0, sizeof(v5dstruct) );

   /* special cases */
   v->Projection = -1;
   v->VerticalSystem = -1;

   for (i=0;i<MAXVARS;i++) {
      v->MinVal[i] = MISSING;
      v->MaxVal[i] = -MISSING;
      v->LowLev[i] = 0;
   }

   /* set file version */
   strcpy(v->FileVersion, FILE_VERSION);

   v->CompressMode = 1;
   v->FileDesc = -1;
}

/*
 * Read a v5d file header.
 * Input:  f - file opened for reading.
 *         v - pointer to v5dstruct to store header info into.
 * Return:  1 = ok, 0 = error.
 */
static int read_v5d_header( v5dstruct *v )
{
#define SKIP(N)   lseek( f, N, SEEK_CUR )
   int end_of_header = 0;
   unsigned int id;
   int idlen, var, numargs;
   int f;

   f = v->FileDesc;

   /* first try to read the header id */
   read_int4( f, (int*) &id );
   read_int4( f, &idlen );
   if (id==TAG_ID && idlen==0) {
      /* this is a v5d file */
      v->FileFormat = 0;
   }
   else if (id>=0x80808080 && id<=0x80808083) {
      /* this is an old COMP* file */
      v->FileFormat = id;
      return read_comp_header( f, v );
   }
   else {
      /* unknown file type */
      printf("Error: not a v5d file\n");
      return 0;
   }

   v->CompressMode = 1; /* default */

   while (!end_of_header) {
      int tag, length;
      int i, var, time, nl, lev;

      if (read_int4(f,&tag)<1 || read_int4(f,&length)<1) {
         printf("Error while reading header, premature EOF\n");
         return 0;
      }

     switch (tag) {
         case TAG_VERSION:
            assert( length==10 );
            read_bytes( f, v->FileVersion, 10 );
            /* Check if reading a file made by a future version of Vis5D */
            if (strcmp(v->FileVersion, FILE_VERSION)>0) {
               printf("Warning: Trying to read a version %s file, you should");
               printf(" upgrade Vis5D.\n");
            }
            break;
         case TAG_NUMTIMES:
            assert( length==4 );
            read_int4( f, &v->NumTimes );
            break;
         case TAG_NUMVARS:
            assert( length==4 );
            read_int4( f, &v->NumVars );
            break;
         case TAG_VARNAME:
            assert( length==14 );   /* 1 int + 10 char */
            read_int4( f, &var );
            read_bytes( f, v->VarName[var], 10 );
            break;
         case TAG_NR:
            /* Number of rows for all variables */
            assert( length==4 );
            read_int4( f, &v->Nr );
            break;
         case TAG_NC:
            /* Number of columns for all variables */
            assert( length==4 );
            read_int4( f, &v->Nc );
            break;
         case TAG_NL:
            /* Number of levels for all variables */
            assert( length==4 );
            read_int4( f, &nl );
            for (i=0;i<v->NumVars;i++) {
               v->Nl[i] = nl;
            }
            break;
         case TAG_NL_VAR:
            /* Number of levels for one variable */
            assert( length==8 );
            read_int4( f, &var );
            read_int4( f, &v->Nl[var] );
            break;
         case TAG_LOWLEV_VAR:
            /* Lowest level for one variable */
            assert( length==8 );
            read_int4( f, &var );
            read_int4( f, &v->LowLev[var] );
            break;

         case TAG_TIME:
            /* Time stamp for 1 timestep */
            assert( length==8 );
            read_int4( f, &time );
            read_int4( f, &v->TimeStamp[time] );
            break;
        case TAG_DATE:
            /* Date stamp for 1 timestep */
            assert( length==8 );
            read_int4( f, &time );
            read_int4( f, &v->DateStamp[time] );
            break;

         case TAG_MINVAL:
            /* Minimum value for a variable */
            assert( length==8 );
            read_int4( f, &var );
            read_float4( f, &v->MinVal[var] );
            break;
         case TAG_MAXVAL:
            /* Maximum value for a variable */
            assert( length==8 );
            read_int4( f, &var );
            read_float4( f, &v->MaxVal[var] );
            break;
         case TAG_COMPRESS:
            /* Compress mode */
            assert( length==4 );
            read_int4( f, &v->CompressMode );
            break;
         case TAG_UNITS:
            /* physical units */
            assert( length==24 );
            read_int4( f, &var );
            read_bytes( f, v->Units[var], 20 );
            break;

         /*
          * Vertical coordinate system
          */
         case TAG_VERTICAL_SYSTEM:
            assert( length==4 );
            read_int4( f, &v->VerticalSystem );
            if (v->VerticalSystem<0 || v->VerticalSystem>3) {
               printf("Error: bad vertical coordinate system: %d\n",
                      v->VerticalSystem );
            }
            break;
         case TAG_VERT_ARGS:
            read_int4( f, &numargs );
            assert( numargs <= MAXVERTARGS );
            read_float4_array( f, v->VertArgs, numargs );
            assert( length==numargs*4+4 );
            break;
         case TAG_HEIGHT:
            /* height of a grid level */
            assert( length==8 );
            read_int4( f, &lev );
            read_float4( f, &v->VertArgs[lev] );
            break;
         case TAG_BOTTOMBOUND:
            assert( length==4 );
            read_float4( f, &v->VertArgs[0] );
            break;
         case TAG_LEVINC:
            assert( length==4 );
            read_float4( f, &v->VertArgs[1] );
            break;
         /*
          * Map projection information
          */
         case TAG_PROJECTION:
            assert( length==4 );
            read_int4( f, &v->Projection );
            if (v->Projection<0 || v->Projection>4) { /* WLH 4-21-95 */
               printf("Error while reading header, bad projection (%d)\n",
                       v->Projection );
               return 0;
            }
            break;
         case TAG_PROJ_ARGS:
            read_int4( f, &numargs );
            assert( numargs <= MAXPROJARGS );
            read_float4_array( f, v->ProjArgs, numargs );
            assert( length==4*numargs+4 );
            break;
         case TAG_NORTHBOUND:
            assert( length==4 );
            if (v->Projection==0 || v->Projection==1 || v->Projection==4) {
               read_float4( f, &v->ProjArgs[0] );
            }
            else {
               SKIP( 4 );
            }
            break;
         case TAG_WESTBOUND:
            assert( length==4 );
            if (v->Projection==0 || v->Projection==1 || v->Projection==4) {
               read_float4( f, &v->ProjArgs[1] );
            }
            else {
               SKIP( 4 );
            }
            break;
         case TAG_ROWINC:
            assert( length==4 );
            if (v->Projection==0 || v->Projection==1 || v->Projection==4) {
               read_float4( f, &v->ProjArgs[2] );
            }
            else {
               SKIP( 4 );
            }
            break;
         case TAG_COLINC:
            assert( length==4 );
            if (v->Projection==0 || v->Projection==1 || v->Projection==4) {
               read_float4( f, &v->ProjArgs[3] );
            }
            else if (v->Projection==2) {
               read_float4( f, &v->ProjArgs[5] );
            }
            else if (v->Projection==3) {
               read_float4( f, &v->ProjArgs[4] );
            }
            else {
               SKIP( 4 );
            }
            break;
         case TAG_LAT1:
            assert( length==4 );
            if (v->Projection==2) {
               read_float4( f, &v->ProjArgs[0] );
            }
            else {
               SKIP( 4 );
            }
            break;
         case TAG_LAT2:
            assert( length==4 );
            if (v->Projection==2) {
               read_float4( f, &v->ProjArgs[1] );
            }
            else {
               SKIP( 4 );
            }
            break;
         case TAG_POLE_ROW:
            assert( length==4 );
            if (v->Projection==2) {
               read_float4( f, &v->ProjArgs[2] );
            }
            else {
               SKIP( 4 );
            }
            break;
         case TAG_POLE_COL:
            assert( length==4 );
            if (v->Projection==2) {
               read_float4( f, &v->ProjArgs[3] );
            }
            else {
               SKIP( 4 );
            }
            break;
         case TAG_CENTLON:
            assert( length==4 );
            if (v->Projection==2) {
               read_float4( f, &v->ProjArgs[4] );
            }
            else if (v->Projection==3) {
               read_float4( f, &v->ProjArgs[1] );
            }
            else if (v->Projection==4) { /* WLH 4-21-95 */
               read_float4( f, &v->ProjArgs[5] );
            }
            else {
               SKIP( 4 );
            }
            break;
         case TAG_CENTLAT:
            assert( length==4 );
            if (v->Projection==3) {
               read_float4( f, &v->ProjArgs[0] );
            }
            else if (v->Projection==4) { /* WLH 4-21-95 */
               read_float4( f, &v->ProjArgs[4] );
            }
            else {
               SKIP( 4 );
            }
            break;
         case TAG_CENTROW:
            assert( length==4 );
            if (v->Projection==3) {
               read_float4( f, &v->ProjArgs[2] );
            }
            else {
               SKIP( 4 );
            }
            break;
         case TAG_CENTCOL:
            assert( length==4 );
            if (v->Projection==3) {
               read_float4( f, &v->ProjArgs[3] );
            }
            else {
               SKIP( 4 );
            }
            break;
         case TAG_ROTATION:
            assert( length==4 );
            if (v->Projection==4) { /* WLH 4-21-95 */
               read_float4( f, &v->ProjArgs[6] );
            }
            else {
               SKIP( 4 );
            }
            break;

         case TAG_END:
            /* end of header */
            end_of_header = 1;
            lseek( f, length, SEEK_CUR );
            break;

         default:
            /* unknown tag, skip to next tag */
            printf("Unknown tag: %d  length=%d\n", tag, length );
            lseek( f, length, SEEK_CUR );
            break;
      }

   }
   v5dVerifyStruct( v );

   /* Now we're ready to read the grid data */

   /* Save current file pointer */
   v->FirstGridPos = ltell(f);

   /* compute grid sizes */
   v->SumGridSizes = 0;
   for (var=0;var<v->NumVars;var++) {
      v->GridSize[var] = 8 * v->Nl[var] + v5dSizeofGrid( v, 0, var );
      v->SumGridSizes += v->GridSize[var];
   }

   return 1;
#undef SKIP
}
/*
 * Do some checking that the information in a v5dstruct is valid.
 * Input:  v - pointer to v5dstruct
 * Return:  1 = g is ok, 0 = g is invalid
 */
int v5dVerifyStruct( const v5dstruct *v )
{
   int var, i, invalid, maxnl;

   invalid = 0;

   if (!v)
      return 0;

   /* Number of variables */
   if (v->NumVars<0) {
      printf("Invalid number of variables: %d\n", v->NumVars );
      invalid = 1;
   }
   else if (v->NumVars>MAXVARS) {
      printf("Too many variables: %d  (Maximum is %d)\n",
             v->NumVars, MAXVARS);
      invalid = 1;
   }

   /* Variable Names */
   for (i=0;i<v->NumVars;i++) {
      if (v->VarName[i][0]==0) {
         printf("Missing variable name: VarName[%d]=\"\"\n", i );
         invalid = 1;
      }
   }

   /* Number of timesteps */
   if (v->NumTimes<0) {
      printf("Invalid number of timesteps: %d\n", v->NumTimes );
      invalid = 1;
   }
   else if (v->NumTimes>MAXTIMES) {
      printf("Too many timesteps: %d  (Maximum is %d)\n",
             v->NumTimes, MAXTIMES );
      invalid = 1;
   }

   /* Make sure timestamps are increasing */
   for (i=1;i<v->NumTimes;i++) {
      int date0 = v5dYYDDDtoDays( v->DateStamp[i-1] );
      int date1 = v5dYYDDDtoDays( v->DateStamp[i] );
      int time0 = v5dHHMMSStoSeconds( v->TimeStamp[i-1] );
      int time1 = v5dHHMMSStoSeconds( v->TimeStamp[i] );
      if (time1<=time0 && date1<=date0) {
         printf("Timestamp for step %d must be later than step %d\n", i, i-1);
         invalid = 1;
      }
   }

   /* Rows */
   if (v->Nr<2) {
      printf("Too few rows: %d (2 is minimum)\n", v->Nr );
      invalid = 1;
   }
   else if (v->Nr>MAXROWS) {
      printf("Too many rows: %d (%d is maximum)\n", v->Nr, MAXROWS );
      invalid = 1;
   }
   /* Columns */
   if (v->Nc<2) {
      printf("Too few columns: %d (2 is minimum)\n", v->Nc );
      invalid = 1;
   }
   else if (v->Nc>MAXCOLUMNS) {
      printf("Too many columns: %d (%d is maximum)\n", v->Nc, MAXCOLUMNS );
      invalid = 1;
   }

   /* Levels */
   maxnl = 0;
   for (var=0;var<v->NumVars;var++) {
      if (v->LowLev[var] < 0) {
         printf("Low level cannot be negative for var %s: %d\n",
                 v->VarName[var], v->LowLev[var] );
         invalid = 1;
      }
      if (v->Nl[var]<1) {
         printf("Too few levels for var %s: %d (1 is minimum)\n",
                 v->VarName[var], v->Nl[var] );
         invalid = 1;
      }
      if (v->Nl[var]+v->LowLev[var]>MAXLEVELS) {
         printf("Too many levels for var %s: %d (%d is maximum)\n",
                 v->VarName[var], v->Nl[var]+v->LowLev[var], MAXLEVELS );
         invalid = 1;
      }
      if (v->Nl[var]+v->LowLev[var]>maxnl) {
         maxnl = v->Nl[var]+v->LowLev[var];
      }
   }

   if (v->CompressMode != 1 && v->CompressMode != 2 && v->CompressMode != 4) {
      printf("Bad CompressMode: %d (must be 1, 2 or 4)\n", v->CompressMode );
      invalid = 1;
   }

   switch (v->VerticalSystem) {
      case 0:
      case 1:
         if (v->VertArgs[1]==0.0) {
            printf("Vertical level increment is zero, must be non-zero\n");
            invalid = 1;
         }
         break;
      case 2:
         /* Check that Height values increase upward */
         for (i=1;i<maxnl;i++) {
            if (v->VertArgs[i] <= v->VertArgs[i-1]) {
               printf("Height[%d]=%f <= Height[%d]=%f, level heights must increase\n",
                      i, v->VertArgs[i], i-1, v->VertArgs[i-1] );
               invalid = 1;
               break;
            }
         }
         break;
      case 3:
         /* Check that Pressure values decrease upward */
         for (i=1;i<maxnl;i++) {
            if (v->VertArgs[i] <= v->VertArgs[i-1]) {
               printf("Pressure[%d]=%f >= Pressure[%d]=%f, level pressures must decrease\n",
                      i, height_to_pressure(v->VertArgs[i]),
                      i-1, height_to_pressure(v->VertArgs[i-1]) );
               invalid = 1;
               break;
            }
         }


         break;
      default:
         printf("VerticalSystem = %d, must be in 0..3\n", v->VerticalSystem );
         invalid = 1;
   }


   switch (v->Projection) {
      case 0:  /* Generic */
         if (v->ProjArgs[2]==0.0) {
            printf("Row Increment (ProjArgs[2]) can't be zero\n");
            invalid = 1;
         }
         if (v->ProjArgs[3]==0.0) {
            printf("Column increment (ProjArgs[3]) can't be zero\n");
            invalid = 1;
         }
         break;
      case 1:  /* Cylindrical equidistant */
         if (v->ProjArgs[2]<0.0) {
            printf("Row Increment (ProjArgs[2]) = %g  (must be >=0.0)\n",
                   v->ProjArgs[2] );
            invalid = 1;
         }
         if (v->ProjArgs[3]<=0.0) {
            printf("Column Increment (ProjArgs[3]) = %g  (must be >=0.0)\n",
                   v->ProjArgs[3] );
            invalid = 1;
         }
         break;
      case 2:  /* Lambert Conformal */
         if (v->ProjArgs[0]<-90.0 || v->ProjArgs[0]>90.0) {
            printf("Lat1 (ProjArgs[0]) out of range: %g\n", v->ProjArgs[0] );
            invalid = 1;
         }
         if (v->ProjArgs[1]<-90.0 || v->ProjArgs[1]>90.0) {
            printf("Lat2 (ProjArgs[1] out of range: %g\n", v->ProjArgs[1] );
            invalid = 1;
         }
         if (v->ProjArgs[5]<=0.0) {
            printf("ColInc (ProjArgs[5]) = %g  (must be >=0.0)\n",
                   v->ProjArgs[5] );
            invalid = 1;
         }
         break;
      case 3:  /* Stereographic */
         if (v->ProjArgs[0]<-90.0 || v->ProjArgs[0]>90.0) {
            printf("Central Latitude (ProjArgs[0]) out of range: ");
            printf("%g  (must be in +/-90)\n", v->ProjArgs[0] );
            invalid = 1;
         }
         if (v->ProjArgs[1]<-180.0 || v->ProjArgs[1]>180.0) {
            printf("Central Longitude (ProjArgs[1]) out of range: ");
            printf("%g  (must be in +/-180)\n", v->ProjArgs[1] );
            invalid = 1;
         }
         if (v->ProjArgs[4]<0) {
            printf("Column spacing (ProjArgs[4]) = %g  (must be positive)\n",
                   v->ProjArgs[4]);
            invalid = 1;
         }
         break;
      case 4:  /* Rotated */
         /* WLH 4-21-95 */
         if (v->ProjArgs[2]<=0.0) {
            printf("Row Increment (ProjArgs[2]) = %g  (must be >=0.0)\n",
                   v->ProjArgs[2] );
            invalid = 1;
         }
         if (v->ProjArgs[3]<=0.0) {
            printf("Column Increment = (ProjArgs[3]) %g  (must be >=0.0)\n",
                   v->ProjArgs[3] );
            invalid = 1;
         }
         if (v->ProjArgs[4]<-90.0 || v->ProjArgs[4]>90.0) {
            printf("Central Latitude (ProjArgs[4]) out of range: ");
            printf("%g  (must be in +/-90)\n", v->ProjArgs[4] );
            invalid = 1;
         }
         if (v->ProjArgs[5]<-180.0 || v->ProjArgs[5]>180.0) {
            printf("Central Longitude (ProjArgs[5]) out of range: ");
            printf("%g  (must be in +/-180)\n", v->ProjArgs[5] );
            invalid = 1;
         }
         if (v->ProjArgs[6]<-180.0 || v->ProjArgs[6]>180.0) {
            printf("Central Longitude (ProjArgs[6]) out of range: ");
            printf("%g  (must be in +/-180)\n", v->ProjArgs[6] );
            invalid = 1;
         }
         break;
      default:
         printf("Projection = %d, must be in 0..4\n", v->Projection );
         invalid = 1;
   }

   return !invalid;
}


/*
 * Read a grid from a v5d file, decompress it and return it.
 * Input:  v - pointer to v5dstruct describing file header
 *         time, var - which timestep and variable.
 *         data - address of buffer to put grid data
 * Output:  data - the grid data
 * Return:  1 = ok, 0 = error.
 */
int v5dReadGrid( v5dstruct *v, int time, int var, float data[] )
{
   float ga[MAXLEVELS], gb[MAXLEVELS];
   void *compdata;
   int bytes;

   if (time<0 || time>=v->NumTimes) {
      printf("Error in v5dReadGrid: bad timestep argument (%d)\n", time);
      return 0;
   }
   if (var<0 || var>=v->NumVars) {
      printf("Error in v5dReadGrid: bad variable argument (%d)\n", var);
      return 0;
   }

   /* allocate compdata buffer */
   if (v->CompressMode==1) {
      bytes = v->Nr * v->Nc * v->Nl[var] * sizeof(unsigned char);
   }
   else if (v->CompressMode==2) {
      bytes = v->Nr * v->Nc * v->Nl[var] * sizeof(unsigned short);
   }
   else if (v->CompressMode==4) {
      bytes = v->Nr * v->Nc * v->Nl[var] * sizeof(float);
   }
   compdata = (void *) malloc( bytes );
   if (!compdata) {
      printf("Error in v5dReadGrid: out of memory (needed %d bytes)\n", bytes);
      return 0;
   }

   /* read the compressed data */
   if (!v5dReadCompressedGrid( v, time, var, ga, gb, compdata )) {
      return 0;
   }

   /* decompress the data */
   v5dDecompressGrid( v->Nr, v->Nc, v->Nl[var], v->CompressMode,
                      compdata, ga, gb, data );

   /* free compdata */
   free( compdata );
   return 1;
}

/*
 * Close a v5d file which was opened with open_v5d_file() or
 * create_v5d_file().
 * Input: f - file descriptor
 * Return:  1 = ok, 0 = error
 */
int v5dCloseFile( v5dstruct *v )
{
   int status = 1;

   if (v->Mode=='w') {
      /* rewrite header because writing grids updates the minval and */
      /* maxval fields */
      lseek( v->FileDesc, 0, SEEK_SET );
      status = write_v5d_header( v );
      lseek( v->FileDesc, 0, SEEK_END );
      close( v->FileDesc );
   }
   else if (v->Mode=='r') {
      /* just close the file */
      close(v->FileDesc);
   }
   else {
      printf("Error in v5dCloseFile: bad v5dstruct argument\n");
      return 0;
   }
   v->FileDesc = -1;
   v->Mode = 0;
   return status;
}


/*
 * Read the header from a COMP* file and return results in the v5dstruct.
 * Input:  f - the file descriptor
 *         v - pointer to a v5dstruct.
 * Return:  1 = ok, 0 = error.
 */
static int read_comp_header( int f, v5dstruct *v )
{
   unsigned int id;

   /* reset file position to start of file */
   lseek( f, 0, SEEK_SET );

   /* read file ID */
   read_int4( f, (int *) &id );

   if (id==0x80808080 || id==0x80808081) {
      /* Older COMP5D format */
      int gridtimes, gridparms;
      int i, j, it, iv, nl;
      int gridsize;
      float hgttop, hgtinc;
      /*char *compgrid;*/

      if (id==0x80808080) {
         /* 20 vars, 300 times */
         gridtimes = 300;
         gridparms = 20;
      }
      else {
         /* 30 vars, 400 times */
         gridtimes = 400;
         gridparms = 30;
      }

      v->FirstGridPos = 12*4 + 8*gridtimes + 4*gridparms;

      read_int4( f, &v->NumTimes );
      read_int4( f, &v->NumVars );
      read_int4( f, &v->Nr );
      read_int4( f, &v->Nc );
      read_int4( f, &nl );
      for (i=0;i<v->NumVars;i++) {
         v->Nl[i] = nl;
         v->LowLev[i] = 0;
      }
      read_float4( f, &v->ProjArgs[0] );
      read_float4( f, &v->ProjArgs[1] );
      read_float4( f, &hgttop );
      read_float4( f, &v->ProjArgs[2] );
      read_float4( f, &v->ProjArgs[3] );
      read_float4( f, &hgtinc );
/*
      for (i=0;i<nl;i++) {
         v->Height[nl-i-1] = hgttop - i * hgtinc;
      }
*/
      v->VerticalSystem = 1;
      v->VertArgs[0] = hgttop - hgtinc * (nl-1);
      v->VertArgs[1] = hgtinc;

      /* read dates and times */
      for (i=0;i<gridtimes;i++) {
         read_int4( f, &j );
         v->DateStamp[i] = v5dDaysToYYDDD( j );
      }
      for (i=0;i<gridtimes;i++) {
         read_int4( f, &j );
         v->TimeStamp[i] = v5dSecondsToHHMMSS( j );
      }

      /* read variable names */
      for (i=0;i<gridparms;i++) {
         char name[4];
         read_bytes( f, name, 4 );
         /* remove trailing spaces, if any */
         for (j=3;j>0;j--) {
            if (name[j]==' ' || name[j]==0)
              name[j] = 0;
            else
              break;
         }
         strncpy( v->VarName[i], name, 4 );
         v->VarName[i][4] = 0;
      }

      gridsize = ( (v->Nr * v->Nc * nl + 3) / 4) * 4;
      for (i=0;i<v->NumVars;i++) {
         v->GridSize[i] = 8 + gridsize;
      }
      v->SumGridSizes = (8+gridsize) * v->NumVars;

      /* read the grids and their ga,gb values to find min and max values */

      for (i=0;i<v->NumVars;i++) {
         v->MinVal[i] = 999999.9;
         v->MaxVal[i] = -999999.9;
      }

      /*compgrid = (char *) malloc( gridsize );*/

      for (it=0; it<v->NumTimes; it++) {
         for (iv=0; iv<v->NumVars; iv++) {
            float ga, gb;
            float min, max;

            read_float4( f, &ga );
            read_float4( f, &gb );

            /* skip ahead by 'gridsize' bytes */
            if (lseek( f, gridsize, SEEK_CUR )==-1) {
               printf("Error:  Unexpected end of file, ");
               printf("file may be corrupted.\n");
               return 0;
            }
            min = -(125.0+gb)/ga;
            max = (125.0-gb)/ga;
            if (min<v->MinVal[iv])  v->MinVal[iv] = min;
            if (max>v->MaxVal[iv])  v->MaxVal[iv] = max;
         }
      }

      /*free( compgrid );*/

      /* done */
   }
   else if (id==0x80808082 || id==0x80808083) {
      /* Newer COMP5D format */
      int gridtimes, gridsize;
      int it, iv, nl, i, j;
      float delta;

      read_int4( f, &gridtimes );
      read_int4( f, &v->NumVars );
      read_int4( f, &v->NumTimes );
      read_int4( f, &v->Nr );
      read_int4( f, &v->Nc );
      read_int4( f, &nl );
      for (i=0;i<v->NumVars;i++) {
         v->Nl[i] = nl;
      }

      read_float4( f, &v->ProjArgs[2] );
      read_float4( f, &v->ProjArgs[3] );

      /* Read height and determine if equal spacing */
      v->VerticalSystem = 1;
      for (i=0;i<nl;i++) {
         read_float4( f, &v->VertArgs[i] );
         if (i==1) {
            delta = v->VertArgs[1] - v->VertArgs[0];
         }
         else if (i>1) {
            if (delta != (v->VertArgs[i] - v->VertArgs[i-1])) {
               v->VerticalSystem = 2;
            }
         }
      }
      if (v->VerticalSystem==1) {
         v->VertArgs[1] = delta;
      }

      /* read variable names */
      for (iv=0; iv<v->NumVars; iv++) {
         char name[8];

         read_bytes( f, name, 8 );

         /* remove trailing spaces, if any */
         for (j=7;j>0;j--) {
            if (name[j]==' ' || name[j]==0)
              name[j] = 0;
            else
              break;
         }
         strncpy( v->VarName[iv], name, 8 );
         v->VarName[iv][8] = 0;
      }

      for (iv=0;iv<v->NumVars;iv++) {
         read_float4( f, &v->MinVal[iv] );
      }
      for (iv=0;iv<v->NumVars;iv++) {
         read_float4( f, &v->MaxVal[iv] );
      }
      for (it=0;it<gridtimes;it++) {
         read_int4( f, &j );
         v->TimeStamp[it] = v5dSecondsToHHMMSS( j );
      }
      for (it=0;it<gridtimes;it++) {
         read_int4( f, &j );
         v->DateStamp[it] = v5dDaysToYYDDD( j );
      }
      for (it=0;it<gridtimes;it++) {
         float nlat;
         read_float4( f, &nlat );
         if (it==0)  v->ProjArgs[0] = nlat;
      }
      for (it=0;it<gridtimes;it++) {
         float wlon;
         read_float4( f, &wlon );
         if (it==0)  v->ProjArgs[1] = wlon;
      }

      /* calculate grid storage sizes */
      if (id==0x80808082) {
         gridsize = nl*2*4 + ( (v->Nr * v->Nc * nl + 3) / 4) * 4;
      }
      else {
         /* McIDAS grid and file numbers present */
         gridsize = 8 + nl*2*4 + ( (v->Nr * v->Nc * nl + 3) / 4) * 4;
      }
      for (i=0;i<v->NumVars;i++) {
         v->GridSize[i] = gridsize;
      }
      v->SumGridSizes = gridsize * v->NumVars;

      /* read McIDAS numbers??? */

      /* size (in bytes) of all header info */
      v->FirstGridPos = 9*4 + v->Nl[0]*4 + v->NumVars*16 + gridtimes*16;

   }

   v->CompressMode = 1; /* one byte per grid point */
   v->Projection = 1;  /* Cylindrical equidistant */
   v->FileVersion[0] = 0;

   return 1;
}

/*
 * Return the size (in bytes) of the 3-D grid specified by time and var.
 * Input:  v - pointer to v5dstruct describing the file
 *         time, var - which timestep and variable
 * Return:  number of data points.
 */
int v5dSizeofGrid( const v5dstruct *v, int time, int var )
{
   return v->Nr * v->Nc * v->Nl[var] * v->CompressMode;
}

/*
 * Convert a date from YYDDD format to days since Jan 1, 1900.
 */
int v5dYYDDDtoDays( int yyddd )
{
   int iy, id, idays;

   iy = yyddd / 1000;
   id = yyddd - 1000*iy;
   if (iy < 50) iy += 100; /* WLH 31 July 96 << 31 Dec 99 */
   idays = 365*iy + (iy-1)/4 + id;

   return idays;
}


/*
 * Convert a time from HHMMSS format to seconds since midnight.
 */
int v5dHHMMSStoSeconds( int hhmmss )
{
   int h, m, s;

   h = hhmmss / 10000;
   m = (hhmmss / 100) % 100;
   s = hhmmss % 100;

   return s + m*60 + h*60*60;
}

float height_to_pressure(float height)
{
  return (float) DEFAULT_LOG_SCALE * exp((double) height / DEFAULT_LOG_EXP);
}

/*
 * Read a compressed grid from a v5d file.
 * Input:  v - pointer to v5dstruct describing the file
 *         time, var - which timestep and variable
 *         ga, gb - arrays to store grid (de)compression values
 *         compdata - address of where to store compressed grid data.
 * Return:  1 = ok, 0 = error.
 */
int v5dReadCompressedGrid( v5dstruct *v, int time, int var,
                           float *ga, float *gb, void *compdata )
{
   int pos, n, k;

   if (time<0 || time>=v->NumTimes) {
      printf("Error in v5dReadCompressedGrid: bad timestep argument (%d)\n",
             time);
      return 0;
   }
   if (var<0 || var>=v->NumVars) {
      printf("Error in v5dReadCompressedGrid: bad var argument (%d)\n",
             var);
      return 0;
   }

   if (v->FileFormat) {
      /* old COMP* file */
      return read_comp_grid( v, time, var, ga, gb, compdata );
   }

   /* move to position in file */
   pos = grid_position( v, time, var );
   lseek( v->FileDesc, pos, SEEK_SET );

   /* read ga, gb arrays */
   read_float4_array( v->FileDesc, ga, v->Nl[var] );
   read_float4_array( v->FileDesc, gb, v->Nl[var] );

   /* read compressed grid data */
   n = v->Nr * v->Nc * v->Nl[var];
   if (v->CompressMode==1) {
      k = read_block( v->FileDesc, compdata, n, 1 )==n;
   }
   else if (v->CompressMode==2) {
      k = read_block( v->FileDesc, compdata, n, 2 )==n;
   }
   else if (v->CompressMode==4) {
      k = read_block( v->FileDesc, compdata, n, 4 )==n;
   }
   if (!k) {
      /* error */
      printf("Error in v5dReadCompressedGrid: read failed, bad file?\n");
   }
   return k;

/*
   n = v->Nr * v->Nc * v->Nl[var] * v->CompressMode;
   if (read( v->FileDesc, compdata, n )==n)
      return 1;
   else
      return 0;
*/
}


/*
 * Decompress a 3-D grid from 1-byte integers to 4-byte floats.
 * Input:  nr, nc, nl - size of grid
 *         compdata - array of [nr*nr*nl*compressmode] bytes
 *         ga, gb - arrays of decompression factors
 *         compressmode - 1, 2 or 4 bytes per grid point
 *         data - address to put decompressed values
 * Output:  data - uncompressed floating point data values
 */
void v5dDecompressGrid( int nr, int nc, int nl, int compressmode,
                        void *compdata, float ga[], float gb[],
                        float data[] )
{
   int nrnc = nr * nc;
   int nrncnl = nr * nc * nl;
   V5Dubyte *compdata1 = (V5Dubyte *) compdata;
   V5Dushort *compdata2 = (V5Dushort *) compdata;

   if (compressmode == 1) {
      int p, i, lev;
      p = 0;
      for (lev=0;lev<nl;lev++) {
         float a = ga[lev];
         float b = gb[lev];

         /* WLH 2-2-95 */
         float d, aa;
         int id;
         if (a > 0.0000000001) {
           d = b / a;
           id = floor(d);
           d = d - id;
           aa = a * 0.000001;
         }
         else {
           id = 1;
         }
         if (-254 <= id && id <= 0 && d < aa) {
           for (i=0;i<nrnc;i++,p++) {
              if (compdata1[p]==255) {
                 data[p] = MISSING;
              }
              else {
                 data[p] = (float) (int) compdata1[p] * a + b;
                 if (fabs(data[p]) < aa) data[p] = aa;
              }
           }
         }
         else {
           for (i=0;i<nrnc;i++,p++) {
              if (compdata1[p]==255) {
                 data[p] = MISSING;
              }
              else {
                 data[p] = (float) (int) compdata1[p] * a + b;
              }
           }
         }
         /* end of WLH 2-2-95 */
      }
   }


   else if (compressmode == 2) {
      int p, i, lev;
      p = 0;
      for (lev=0;lev<nl;lev++) {
         float a = ga[lev];
         float b = gb[lev];
#ifdef _CRAY
         /* this is tricky because sizeof(V5Dushort)==8, not 2 */
         for (i=0;i<nrnc;i++,p++) {
            int compvalue;
            compvalue = (compdata1[p*2] << 8) | compdata1[p*2+1];
            if (compvalue==65535) {
               data[p] = MISSING;
            }
            else {
               data[p] = (float) compvalue * a + b;
            }
         }
#else
         /* sizeof(V5Dushort)==2! */
         for (i=0;i<nrnc;i++,p++) {
            if (compdata2[p]==65535) {
               data[p] = MISSING;
            }
            else {
               data[p] = (float) (int) compdata2[p] * a + b;
            }
         }
#endif
      }
   }

   else {
      /* compressmode==4 */
#ifdef _CRAY
      ieee_to_cray_array( data, compdata, nrncnl );
#else
      /* other machines: just copy 4-byte IEEE floats */
      assert( sizeof(float)==4 );
      memcpy( data, compdata, nrncnl*4 );
#endif
   }
}
/*
 * Write the information in the given v5dstruct as a v5d file header.
 * Note that the current file position is restored when this function
 * returns normally.
 * Input:  f - file already open for writing
 *         v - pointer to v5dstruct
 * Return:  1 = ok, 0 = error.
 */
static int write_v5d_header( v5dstruct *v )
{
   int var, time, filler, maxnl;
   int f;
   int newfile;

   if (v->FileFormat!=0) {
      printf("Error: v5d library can't write comp5d format files.\n");
      return 0;
   }

   f = v->FileDesc;

   if (!v5dVerifyStruct( v ))
      return 0;

   /* Determine if we're writing to a new file */
   if (v->FirstGridPos==0) {
      newfile = 1;
   }
   else {
      newfile = 0;
   }

   /* compute grid sizes */
   v->SumGridSizes = 0;
   for (var=0;var<v->NumVars;var++) {
      v->GridSize[var] = 8 * v->Nl[var] + v5dSizeofGrid( v, 0, var );
      v->SumGridSizes += v->GridSize[var];
   }

   /* set file pointer to start of file */
   lseek( f, 0, SEEK_SET );
   v->CurPos = 0;

   /*
    * Write the tagged header info
    */
#define WRITE_TAG( V, T, L )  if (!write_tag(V,T,L,newfile))  return 0;

   /* ID */
   WRITE_TAG( v, TAG_ID, 0 );

   /* File Version */
   WRITE_TAG( v, TAG_VERSION, 10 );
   write_bytes( f, FILE_VERSION, 10 );

   /* Number of timesteps */
   WRITE_TAG( v, TAG_NUMTIMES, 4 );
   write_int4( f, v->NumTimes );

   /* Number of variables */
   WRITE_TAG( v, TAG_NUMVARS, 4 );
   write_int4( f, v->NumVars );

  /* Names of variables */
   for (var=0;var<v->NumVars;var++) {
      WRITE_TAG( v, TAG_VARNAME, 14 );
      write_int4( f, var );
      write_bytes( f, v->VarName[var], 10 );
   }

   /* Physical Units */
   for (var=0;var<v->NumVars;var++) {
      WRITE_TAG( v, TAG_UNITS, 24 );
      write_int4( f, var );
      write_bytes( f, v->Units[var], 20 );
   }

   /* Date and time of each timestep */
   for (time=0;time<v->NumTimes;time++) {
      WRITE_TAG( v, TAG_TIME, 8 );
      write_int4( f, time );
      write_int4( f, v->TimeStamp[time] );
      WRITE_TAG( v, TAG_DATE, 8 );
      write_int4( f, time );
      write_int4( f, v->DateStamp[time] );
   }

   /* Number of rows */
   WRITE_TAG( v, TAG_NR, 4 );
   write_int4( f, v->Nr );

   /* Number of columns */
   WRITE_TAG( v, TAG_NC, 4 );
   write_int4( f, v->Nc );

   /* Number of levels, compute maxnl */
   maxnl = 0;
   for (var=0;var<v->NumVars;var++) {
      WRITE_TAG( v, TAG_NL_VAR, 8 );
      write_int4( f, var );
      write_int4( f, v->Nl[var] );
      WRITE_TAG( v, TAG_LOWLEV_VAR, 8 );
      write_int4( f, var );
      write_int4( f, v->LowLev[var] );
      if (v->Nl[var]+v->LowLev[var]>maxnl) {
         maxnl = v->Nl[var]+v->LowLev[var];
      }
   }

   /* Min/Max values */
   for (var=0;var<v->NumVars;var++) {
      WRITE_TAG( v, TAG_MINVAL, 8 );
      write_int4( f, var );
      write_float4( f, v->MinVal[var] );
      WRITE_TAG( v, TAG_MAXVAL, 8 );
      write_int4( f, var );
      write_float4( f, v->MaxVal[var] );
   }

   /* Compress mode */
   WRITE_TAG( v, TAG_COMPRESS, 4 );
   write_int4( f, v->CompressMode );

   WRITE_TAG( v, TAG_VERTICAL_SYSTEM, 4 );
   write_int4( f, v->VerticalSystem );
   WRITE_TAG( v, TAG_VERT_ARGS, 4+4*MAXVERTARGS );
   write_int4( f, MAXVERTARGS );
   write_float4_array( f, v->VertArgs, MAXVERTARGS );

   /* Map Projection */
   WRITE_TAG( v, TAG_PROJECTION, 4 );
   write_int4( f, v->Projection );
   WRITE_TAG( v, TAG_PROJ_ARGS, 4+4*MAXPROJARGS );
   write_int4( f, MAXPROJARGS );
   write_float4_array( f, v->ProjArgs, MAXPROJARGS );

   /* write END tag */
   if (newfile) {
      /* We're writing to a brand new file.  Reserve 10000 bytes */
      /* for future header growth. */
      WRITE_TAG( v, TAG_END, 10000 );
      lseek( f, 10000, SEEK_CUR );

      /* Let file pointer indicate where first grid is stored */
      v->FirstGridPos = ltell( f );
   }
   else {
      /* we're rewriting a header */
      filler = v->FirstGridPos - ltell(f);
      WRITE_TAG( v, TAG_END, filler-8 );
   }

#undef WRITE_TAG

   return 1;
}

/*
 * Convert a day since Jan 1, 1900 to YYDDD format.
 */
int v5dDaysToYYDDD( int days )
{
   int iy, id, iyyddd;

   iy = (4*days)/1461;
   id = days-(365*iy+(iy-1)/4);
   if (iy > 99) iy = iy - 100; /* WLH 31 July 96 << 31 Dec 99 */
   /* iy = iy + 1900; is the right way to fix this, but requires
      changing all places where dates are printed - procrastinate */
   iyyddd = iy*1000+id;

   return iyyddd;
}

/*
 * Convert a time in seconds since midnight to HHMMSS format.
 */
int v5dSecondsToHHMMSS( int seconds )
{
   int hh, mm, ss;

   hh = seconds / (60*60);
   mm = (seconds / 60) % 60;
   ss = seconds % 60;
   return hh*10000 + mm * 100 + ss;
}

/*
 * Read a compressed grid from a COMP* file.
 * Return:  1 = ok, 0 = error.
 */
static int read_comp_grid( v5dstruct *v, int time, int var,
                           float *ga, float *gb, void *compdata )
{
   unsigned int pos;
   V5Dubyte bias;
   int i, n, nl;
   int f;
   V5Dubyte *compdata1 = (V5Dubyte *) compdata;

   f = v->FileDesc;

   /* move to position in file */
   pos = grid_position( v, time, var );
   lseek( f, pos, SEEK_SET );

   if (v->FileFormat==0x80808083) {
      /* read McIDAS grid and file numbers */
      int mcfile, mcgrid;
      read_int4( f, &mcfile );
      read_int4( f, &mcgrid );
      v->McFile[time][var] = (short) mcfile;
      v->McGrid[time][var] = (short) mcgrid;
   }

   nl = v->Nl[var];

   if (v->FileFormat==0x80808080 || v->FileFormat==0x80808081) {
      /* single ga,gb pair for whole grid */
      float a, b;
      read_float4( f, &a );
      read_float4( f, &b );
      /* convert a, b to new v5d ga, gb values */
      for (i=0;i<nl;i++) {
         if (a==0.0) {
            ga[i] = gb[i] = 0.0;
         }
         else {
            gb[i] = (b+128.0) / -a;
            ga[i] = 1.0 / a;
         }
      }
      bias = 128;
   }
   else {
      /* read ga, gb arrays */
      read_float4_array( f, ga, v->Nl[var] );
      read_float4_array( f, gb, v->Nl[var] );

      /* convert ga, gb values to v5d system */
      for (i=0;i<nl;i++) {
         if (ga[i]==0.0) {
            ga[i] = gb[i] = 0.0;
         }
         else {
            /*gb[i] = (gb[i]+125.0) / -ga[i];*/
            gb[i] = (gb[i]+128.0) / -ga[i];
            ga[i] = 1.0 / ga[i];
         }
      }
      bias = 128;  /* 125 ??? */
   }
   /* read compressed grid data */
   n = v->Nr * v->Nc * v->Nl[var];
   if (read_bytes( f, compdata1, n )!=n)
      return 0;

   /* convert data values to v5d system */
   n = v->Nr * v->Nc * v->Nl[var];
   for (i=0;i<n;i++) {
      compdata1[i] += bias;
   }

   return 1;
}

/*
 * Compute the location of a compressed grid within a file.
 * Input:  v - pointer to v5dstruct describing the file header.
 *         time, var - which timestep and variable.
 * Return:  file offset in bytes
 */
static int grid_position( const v5dstruct *v, int time, int var )
{
   int pos, i;

   assert( time >= 0 );
   assert( var >= 0 );
   assert( time < v->NumTimes );
   assert( var < v->NumVars );

   pos = v->FirstGridPos + time * v->SumGridSizes;
   for (i=0;i<var;i++) {
      pos += v->GridSize[i];
   }

   return pos;
}

static int write_tag( v5dstruct *v, int tag, int length, int newfile )
{
   if (!newfile) {
      /* have to check that there's room in header to write this tagged item */
      if (v->CurPos+8+length > v->FirstGridPos) {
         printf("Error: out of header space!\n");
         /* Out of header space! */
         return 0;
      }
   }

   if (write_int4( v->FileDesc, tag )==0)  return 0;
   if (write_int4( v->FileDesc, length )==0)  return 0;
   v->CurPos += 8 + length;
   return 1;
}

